# -*- coding: utf-8 -*-

import json

from .api_request import APIRequest
from ..objects import PublicResourcesListObject, TrashResourceObject
from ..objects import FilesResourceListObject, LastUploadedResourceListObject
from ..objects import ResourceObject, ResourceUploadLinkObject, PublicResourceObject
from ..objects import OperationLinkObject, ResourceLinkObject, ResourceDownloadLinkObject
from ..common import is_operation_link, ensure_path_has_schema
from ..exceptions import InvalidResponseError

from ..compat import Iterable, Dict, List
from typing import Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:
    import aiohttp
    from ..yadisk import YaDisk

__all__ = ["GetPublicResourcesRequest", "UnpublishRequest", "GetDownloadLinkRequest",
           "GetTrashRequest", "RestoreTrashRequest", "DeleteTrashRequest",
           "LastUploadedRequest", "CopyRequest", "GetMetaRequest", "GetUploadLinkRequest",
           "MkdirRequest", "PublishRequest", "UploadURLRequest", "DeleteRequest",
           "SaveToDiskRequest", "GetPublicMetaRequest", "GetPublicDownloadLinkRequest",
           "MoveRequest", "FilesRequest", "PatchRequest"]

Fields = Iterable[str]

def _substitute_keys(keys: Iterable[str], sub_map: Dict[str, str]) -> List[str]:
    return [".".join(sub_map.get(f, f) for f in k.split(".")) for k in keys]

class GetPublicResourcesRequest(APIRequest):
    """
        A request to get a list of public resources.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param offset: offset from the beginning of the list
        :param limit: maximum number of elements in the list
        :param preview_size: size of the file preview
        :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
        :param type: filter based on type of resources ("file" or "dir")
        :param fields: list of keys to be included in the response

        :returns: :any:`PublicResourcesListObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/resources/public"
    method = "GET"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 offset: int = 0,
                 limit: int = 20,
                 preview_size: Optional[str] = None,
                 preview_crop: Optional[bool] = None,
                 type: Optional[str] = None,
                 fields: Optional[Iterable[str]] = None, **kwargs):
        APIRequest.__init__(self, session, {"offset":       offset,
                                            "limit":        limit,
                                            "preview_size": preview_size,
                                            "preview_crop": preview_crop,
                                            "type":         type,
                                            "fields":       fields}, **kwargs)

    def process_args(self,
                     offset: int,
                     limit: int,
                     preview_size: Optional[str],
                     preview_crop: Optional[bool],
                     type: Optional[str],
                     fields: Optional[Iterable[str]]) -> None:
        self.params["offset"] = offset
        self.params["limit"] = limit

        if preview_size is not None:
            self.params["preview_size"] = preview_size

        if preview_crop is not None:
            self.params["preview_crop"] = preview_crop

        if type is not None:
            self.params["type"] = type

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> PublicResourcesListObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return PublicResourcesListObject(js, yadisk)

class UnpublishRequest(APIRequest):
    """
        A request to make a public resource private.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param path: path to the resource to be unpublished
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceLinkObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/resources/unpublish"
    method = "PUT"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 path: str,
                 fields: Optional[Iterable[str]] = None, **kwargs):
        APIRequest.__init__(self, session, {"path":   path,
                                            "fields": fields}, **kwargs)

    def process_args(self, path: str, fields: Optional[Iterable[str]]) -> None:
        self.params["path"] = ensure_path_has_schema(path)

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> ResourceLinkObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return ResourceLinkObject(js, yadisk)

class GetDownloadLinkRequest(APIRequest):
    """
        A request to get a download link to a resource.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param path: path to the resource to be downloaded
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceDownloadLinkObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/resources/download"
    method = "GET"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 path: str,
                 fields: Optional[Iterable[str]] = None, **kwargs):
        APIRequest.__init__(
            self, session, {"path": path, "fields": fields}, **kwargs)

    def process_args(self, path: str, fields: Optional[Iterable[str]]) -> None:
        self.params["path"] = ensure_path_has_schema(path)

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> ResourceDownloadLinkObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return ResourceDownloadLinkObject(js, yadisk)

class GetTrashRequest(APIRequest):
    """
        A request to get meta-information about a trash resource.

        :param path: path to the trash resource
        :param limit: number of children resources to be included in the response
        :param offset: number of children resources to be skipped in the response
        :param preview_size: size of the file preview
        :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
        :param sort: `str`, field to be used as a key to sort children resources
        :param fields: list of keys to be included in the response

        :returns: :any:`TrashResourceObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/trash/resources"
    method = "GET"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 path: str,
                 offset: int = 0,
                 limit: int = 20,
                 sort: Optional[str] = None,
                 preview_size: Optional[str] = None,
                 preview_crop: Optional[bool] = None,
                 fields: Optional[Iterable[str]] = None, **kwargs):
        APIRequest.__init__(self, session, {"path":         path,
                                            "offset":       offset,
                                            "limit":        limit,
                                            "sort":         sort,
                                            "preview_size": preview_size,
                                            "preview_crop": preview_crop,
                                            "fields":       fields}, **kwargs)

    def process_args(self,
                     path: str,
                     offset: int,
                     limit: int,
                     sort: Optional[str],
                     preview_size: Optional[str],
                     preview_crop: Optional[bool],
                     fields: Optional[Iterable[str]]) -> None:
        self.params["path"] = ensure_path_has_schema(path, "trash")
        self.params["offset"] = offset
        self.params["limit"] = limit

        if sort is not None:
            self.params["sort"] = sort

        if preview_size is not None:
            self.params["preview_size"] = preview_size

        if preview_crop is not None:
            self.params["preview_crop"] = preview_crop

        if fields is not None:
            sub_map = {"embedded": "_embedded"}

            self.params["fields"] = ",".join(_substitute_keys(fields, sub_map))

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> TrashResourceObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return TrashResourceObject(js, yadisk)

class RestoreTrashRequest(APIRequest):
    """
        A request to restore trash.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param path: path to the trash resource to be restored
        :param dst_path: destination path
        :param force_async: forces the operation to be executed asynchronously
        :param overwrite: `bool`, determines whether the destination can be overwritten
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceLinkObject` or :any:`OperationLinkObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/trash/resources/restore"
    method = "PUT"
    success_codes = {201, 202}

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 path: str,
                 dst_path: Optional[str] = None,
                 force_async: bool = False,
                 overwrite: bool = False,
                 fields: Optional[Iterable[str]] = None, **kwargs):
        APIRequest.__init__(self, session, {"path":        path,
                                            "dst_path":    dst_path,
                                            "overwrite":   overwrite,
                                            "force_async": force_async,
                                            "fields":      fields}, **kwargs)

    def process_args(self,
                     path: str,
                     dst_path: Optional[str],
                     force_async: bool,
                     overwrite: bool,
                     fields: Optional[Iterable[str]]) -> None:
        self.params["path"] = ensure_path_has_schema(path, "trash")
        self.params["overwrite"] = "true" if overwrite else "false"
        self.params["force_async"] = "true" if force_async else "false"

        if dst_path is not None:
            self.params["name"] = dst_path

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> Union[OperationLinkObject, ResourceLinkObject]:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        if is_operation_link(js.get("href", "")):
            return OperationLinkObject(js, yadisk)

        return ResourceLinkObject(js, yadisk)

class DeleteTrashRequest(APIRequest):
    """
        A request to delete a trash resource.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param path: path to the trash resource to be deleted
        :param force_async: forces the operation to be executed asynchronously
        :param fields: list of keys to be included in the response

        :returns: :any:`OperationLinkObject` or `None`
    """

    url = "https://cloud-api.yandex.net/v1/disk/trash/resources"
    method = "DELETE"
    success_codes = {202, 204}

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 path: Optional[str] = None,
                 force_async: bool = False,
                 fields: Optional[Iterable[str]] = None, **kwargs):
        APIRequest.__init__(self, session, {"path":        path,
                                            "force_async": force_async,
                                            "fields":      fields}, **kwargs)

    def process_args(self,
                     path: Optional[str],
                     force_async: bool,
                     fields: Optional[Iterable[str]]) -> None:
        if path is not None:
            self.params["path"] = ensure_path_has_schema(path, "trash")

        self.params["force_async"] = "true" if force_async else "false"

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> Optional[OperationLinkObject]:
        if js is not None:
            return OperationLinkObject(js, yadisk)

class LastUploadedRequest(APIRequest):
    """
        A request to get the list of latest uploaded files sorted by upload date.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param limit: maximum number of elements in the list
        :param media_type: type of files to include in the list
        :param preview_size: size of the file preview
        :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
        :param fields: list of keys to be included in the response

        :returns: :any:`LastUploadedResourceListObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/resources/last-uploaded"
    method = "GET"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 limit: int = 20,
                 media_type: Optional[Union[str, Iterable[str]]] = None,
                 preview_size: Optional[str] = None,
                 preview_crop: Optional[bool] = None,
                 fields: Optional[Iterable[str]] = None, **kwargs):
        APIRequest.__init__(self, session, {"limit":        limit,
                                            "media_type":   media_type,
                                            "preview_size": preview_size,
                                            "preview_crop": preview_crop,
                                            "fields":       fields}, **kwargs)

    def process_args(self,
                     limit: int,
                     media_type: Optional[str],
                     preview_size: Optional[str],
                     preview_crop: Optional[bool],
                     fields: Optional[Iterable[str]]) -> None:
        self.params["limit"] = limit

        if media_type is not None:
            if not isinstance(media_type, Iterable):
                raise TypeError("media_type should be a string or an iterable")

            if isinstance(media_type, str):
                self.params["media_type"] = media_type
            else:
                self.params["media_type"] = ",".join(media_type)

        if preview_size is not None:
            self.params["preview_size"] = preview_size

        if preview_crop is not None:
            self.params["preview_crop"] = preview_crop

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> LastUploadedResourceListObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")
        return LastUploadedResourceListObject(js, yadisk)

class CopyRequest(APIRequest):
    """
        A request to copy a file or a directory.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param src_path: source path
        :param dst_path: destination path
        :param overwrite: if `True` the destination path can be overwritten,
                          otherwise, an error will be raised
        :param force_async: forces the operation to be executed asynchronously
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceLinkObject` or :any:`OperationLinkObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/resources/copy"
    method = "POST"
    success_codes = {201, 202}

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 src_path: str,
                 dst_path: str,
                 overwrite: bool = False,
                 force_async: bool = False,
                 fields: Optional[Fields] = None, **kwargs):
        APIRequest.__init__(self, session, {"src_path":    src_path,
                                            "dst_path":    dst_path,
                                            "overwrite":   overwrite,
                                            "force_async": force_async,
                                            "fields":      fields}, **kwargs)

    def process_args(self,
                     src_path: str,
                     dst_path: str,
                     overwrite: bool,
                     force_async: bool,
                     fields: Optional[Fields]) -> None:
        self.params["from"] = ensure_path_has_schema(src_path)
        self.params["path"] = ensure_path_has_schema(dst_path)
        self.params["overwrite"] = "true" if overwrite else "false"
        self.params["force_async"] = "true" if force_async else "false"

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> Union[OperationLinkObject, ResourceLinkObject]:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        if is_operation_link(js.get("href", "")):
            return OperationLinkObject(js, yadisk)

        return ResourceLinkObject(js, yadisk)

class GetMetaRequest(APIRequest):
    """
        A request to get meta-information about a resource.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param path: path to the resource
        :param limit: number of children resources to be included in the response
        :param offset: number of children resources to be skipped in the response
        :param preview_size: size of the file preview
        :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
        :param sort: `str`, field to be used as a key to sort children resources
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/resources"
    method = "GET"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 path: str,
                 limit: Optional[int] = None,
                 offset: Optional[int] = None,
                 preview_size: Optional[str] = None,
                 preview_crop: Optional[bool] = None,
                 sort: Optional[str] = None,
                 fields: Optional[Fields] = None, **kwargs):
        APIRequest.__init__(self, session,
                            {"path":         path,
                             "limit":        limit,
                             "offset":       offset,
                             "preview_size": preview_size,
                             "preview_crop": preview_crop,
                             "sort":         sort,
                             "fields":       fields}, **kwargs)

    def process_args(self,
                     path: str,
                     limit: Optional[int],
                     offset: Optional[int],
                     preview_size: Optional[str],
                     preview_crop: Optional[bool],
                     sort: Optional[str],
                     fields: Optional[Fields]) -> None:
        self.params["path"] = ensure_path_has_schema(path)

        if limit is not None:
            self.params["limit"] = limit

        if offset is not None:
            self.params["offset"] = offset

        if preview_size is not None:
            self.params["preview_size"] = preview_size

        if preview_crop is not None:
            self.params["preview_crop"] = "true" if preview_crop else "false"

        if sort is not None:
            self.params["sort"] = sort

        if fields is not None:
            sub_map = {"embedded": "_embedded"}

            self.params["fields"] = ",".join(_substitute_keys(fields, sub_map))

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> ResourceObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return ResourceObject(js, yadisk)

class GetUploadLinkRequest(APIRequest):
    """
        A request to get an upload link.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param path: path to be uploaded at
        :param overwrite: `bool`, determines whether to overwrite the destination
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceUploadLinkObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/resources/upload"
    method = "GET"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 path: str,
                 overwrite: bool = False,
                 fields: Optional[Fields] = None, **kwargs):
        APIRequest.__init__(self, session, {"path":      path,
                                            "overwrite": overwrite,
                                            "fields":    fields}, **kwargs)

    def process_args(self, path: str, overwrite: bool, fields: Optional[Fields]) -> None:
        self.params["path"] = ensure_path_has_schema(path)
        self.params["overwrite"] = "true" if overwrite else "false"

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> ResourceUploadLinkObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return ResourceUploadLinkObject(js, yadisk)

class MkdirRequest(APIRequest):
    """
        A request to create a new directory.

        :param path: path to the directory to be created
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceLinkObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/resources"
    method = "PUT"
    success_codes = {201}

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 path: str,
                 fields: Optional[Fields] = None, **kwargs):
        APIRequest.__init__(self, session, {"path": path, "fields": fields},
                            **kwargs)

    def process_args(self, path: str, fields: Optional[Fields]) -> None:
        self.params["path"] = ensure_path_has_schema(path)

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> ResourceLinkObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return ResourceLinkObject(js, yadisk)

class PublishRequest(APIRequest):
    """
        A request to make a resource public.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param path: path to the resource to be published
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceLinkObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/resources/publish"
    method = "PUT"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 path: str,
                 fields: Optional[Fields] = None, **kwargs):
        APIRequest.__init__(self, session, {"path":   path,
                                            "fields": fields}, **kwargs)

    def process_args(self, path: str, fields: Optional[Fields]) -> None:
        self.params["path"] = ensure_path_has_schema(path)

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> ResourceLinkObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return ResourceLinkObject(js, yadisk)

class UploadURLRequest(APIRequest):
    """
        A request to upload a file from URL.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param url: source URL
        :param path: destination path
        :param disable_redirects: `bool`, forbid redirects
        :param fields: list of keys to be included in the response

        :returns: :any:`OperationLinkObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/resources/upload"
    method = "POST"
    success_codes = {202}

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 url: str,
                 path: str,
                 disable_redirects: bool = False,
                 fields: Optional[Fields] = None, **kwargs):
        APIRequest.__init__(self, session, {"url":               url,
                                            "path":              path,
                                            "disable_redirects": disable_redirects,
                                            "fields":            fields}, **kwargs)

    def process_args(self,
                     url: str,
                     path: str,
                     disable_redirects: bool,
                     fields: Optional[Fields]) -> None:
        self.params["url"] = url
        self.params["path"] = ensure_path_has_schema(path)
        self.params["disable_redirects"] = "true" if disable_redirects else "false"

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> OperationLinkObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return OperationLinkObject(js, yadisk)

class DeleteRequest(APIRequest):
    """
        A request to delete a file or a directory.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param path: path to the resource to be removed
        :param permanently: if `True`, the resource will be removed permanently,
                            otherwise, it will be just moved to the trash
        :param force_async: forces the operation to be executed asynchronously
        :param md5: `str`, MD5 hash of the file to remove
        :param fields: list of keys to be included in the response

        :returns: :any:`OperationLinkObject` or `None`
    """

    url = "https://cloud-api.yandex.net/v1/disk/resources"
    method = "DELETE"
    success_codes = {202, 204}

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 path: str,
                 permanently: bool = False,
                 md5: Optional[str] = None,
                 force_async: bool = False,
                 fields: Optional[Fields] = None, **kwargs):
        APIRequest.__init__(self, session, {"path":        path,
                                            "permanently": permanently,
                                            "md5":         md5,
                                            "force_async": force_async,
                                            "fields":      fields}, **kwargs)

    def process_args(self,
                     path: str,
                     permanently: bool,
                     md5: Optional[str],
                     force_async: bool,
                     fields: Optional[Fields]) -> None:
        self.params["path"] = ensure_path_has_schema(path)
        self.params["permanently"] = "true" if permanently else "false"
        self.params["force_async"] = "true" if force_async else "false"

        if md5 is not None:
            self.params["md5"] = md5

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> Optional[OperationLinkObject]:
        if js is not None:
            return OperationLinkObject(js, yadisk)

class SaveToDiskRequest(APIRequest):
    """
        A request to save a public resource to the disk.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param public_key: public key or public URL of the resource
        :param name: filename of the saved resource
        :param path: path to the copied resource in the public folder
        :param save_path: path to the destination directory (downloads directory by default)
        :param force_async: forces the operation to be executed asynchronously
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceLinkObject` or :any:`OperationLinkObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/public/resources/save-to-disk"
    method = "POST"
    success_codes = {201, 202}

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 public_key: str,
                 name: Optional[str] = None,
                 path: Optional[str] = None,
                 save_path: Optional[str] = None,
                 force_async: bool = False,
                 fields: Optional[Fields] = None, **kwargs):
        APIRequest.__init__(self, session, {"public_key":  public_key,
                                            "name":        name,
                                            "path":        path,
                                            "save_path":   save_path,
                                            "force_async": force_async,
                                            "fields":      fields}, **kwargs)

    def process_args(self,
                     public_key: str,
                     name: Optional[str],
                     path: Optional[str],
                     save_path: Optional[str],
                     force_async: bool,
                     fields: Optional[Fields]) -> None:
        self.params["public_key"] = public_key

        if name is not None:
            self.params["name"] = name

        if path is not None:
            self.params["path"] = path

        if save_path is not None:
            self.params["save_path"] = ensure_path_has_schema(save_path)

        self.params["force_async"] = "true" if force_async else "false"

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> Union[OperationLinkObject, ResourceLinkObject]:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        if is_operation_link(js.get("href", "")):
            return OperationLinkObject(js, yadisk)

        return ResourceLinkObject(js, yadisk)

class GetPublicMetaRequest(APIRequest):
    """
        A request to get meta-information about a public resource.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param public_key: public key or public URL of the resource
        :param path: relative path to a resource in a public folder.
                     By specifying the key of the published folder in `public_key`,
                     you can request metainformation for any resource in the folder.
        :param offset: offset from the beginning of the list of nested resources
        :param limit: maximum number of nested elements to be included in the list
        :param sort: `str`, field to be used as a key to sort children resources
        :param preview_size: file preview size
        :param preview_crop: `bool`, allow preview crop
        :param fields: list of keys to be included in the response

        :returns: :any:`PublicResourceObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/public/resources"
    method = "GET"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 public_key: str,
                 offset: int = 0,
                 limit: int = 20,
                 path: Optional[str] = None,
                 sort: Optional[str] = None,
                 preview_size: Optional[str] = None,
                 preview_crop: Optional[bool] = None,
                 fields: Optional[Fields] = None, **kwargs):
        APIRequest.__init__(self, session, {"public_key":   public_key,
                                            "offset":       offset,
                                            "limit":        limit,
                                            "path":         path,
                                            "sort":         sort,
                                            "preview_size": preview_size,
                                            "preview_crop": preview_crop,
                                            "fields":       fields}, **kwargs)

    def process_args(self,
                     public_key: str,
                     offset: int,
                     limit: int,
                     path: Optional[str],
                     sort: Optional[str],
                     preview_size: Optional[str],
                     preview_crop: Optional[bool],
                     fields: Optional[Fields]) -> None:
        self.params["public_key"] = public_key
        self.params["offset"] = offset
        self.params["limit"] = limit

        if path is not None:
            self.params["path"] = path

        if sort is not None:
            self.params["sort"] = sort

        if preview_size is not None:
            self.params["preview_size"] = preview_size

        if preview_crop is not None:
            self.params["preview_crop"] = preview_crop

        if fields is not None:
            sub_map = {"embedded": "_embedded",
                       "view_count": "views_count"}

            self.params["fields"] = ",".join(_substitute_keys(fields, sub_map))

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> PublicResourceObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return PublicResourceObject(js, yadisk)

class GetPublicDownloadLinkRequest(APIRequest):
    """
        A request to get a download link for a public resource.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param public_key: public key or public URL of the resource
        :param path: relative path to the resource within the public folder
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceDownloadLinkObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/public/resources/download"
    method = "GET"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 public_key: str,
                 path: Optional[str] = None,
                 fields: Optional[Fields] = None, **kwargs):
        APIRequest.__init__(self, session, {"public_key": public_key,
                                            "path":       path,
                                            "fields":     fields}, **kwargs)

    def process_args(self,
                     public_key: str,
                     path: Optional[str],
                     fields: Optional[Fields]) -> None:
        self.params["public_key"] = public_key

        if path is not None:
            self.params["path"] = path;

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> ResourceDownloadLinkObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return ResourceDownloadLinkObject(js, yadisk)

class MoveRequest(APIRequest):
    """
        A request to move a resource.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param src_path: source path to be moved
        :param dst_path: destination path
        :param force_async: forces the operation to be executed asynchronously
        :param overwrite: `bool`, determines whether to overwrite the destination
        :param fields: list of keys to be included in the response

        :returns: :any:`OperationLinkObject` or :any:`ResourceLinkObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/resources/move"
    method = "POST"
    success_codes = {201, 202}

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 src_path: str,
                 dst_path: str,
                 force_async: bool = False,
                 overwrite: bool = False,
                 fields: Optional[Fields] = None, **kwargs):
        APIRequest.__init__(self, session, {"src_path":    src_path,
                                            "dst_path":    dst_path,
                                            "force_async": force_async,
                                            "overwrite":   overwrite,
                                            "fields":      fields}, **kwargs)

    def process_args(self,
                     src_path: str,
                     dst_path: str,
                     force_async: bool,
                     overwrite: bool,
                     fields: Optional[Fields]) -> None:
        self.params["from"] = ensure_path_has_schema(src_path)
        self.params["path"] = ensure_path_has_schema(dst_path)
        self.params["overwrite"] = "true" if overwrite else "false"
        self.params["force_async"] = "true" if force_async else "false"

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> Union[OperationLinkObject, ResourceLinkObject]:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        if is_operation_link(js.get("href", "")):
            return OperationLinkObject(js, yadisk)

        return ResourceLinkObject(js, yadisk)

class FilesRequest(APIRequest):
    """
        A request to get a flat list of all files (that doesn't include directories).

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param offset: offset from the beginning of the list
        :param limit: number of list elements to be included
        :param media_type: type of files to include in the list
        :param sort: `str`, field to be used as a key to sort children resources
        :param preview_size: size of the file preview
        :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
        :param fields: list of keys to be included in the response

        :returns: :any:`FilesResourceListObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/resources/files"
    method = "GET"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 offset: int = 0,
                 limit: int = 20,
                 media_type: Optional[Union[str, Iterable[str]]] = None,
                 preview_size: Optional[str] = None,
                 preview_crop: Optional[bool] = None,
                 sort: Optional[str] = None,
                 fields: Optional[Fields] = None, **kwargs):
        APIRequest.__init__(self, session, {"offset":       offset,
                                            "limit":        limit,
                                            "media_type":   media_type,
                                            "sort":         sort,
                                            "preview_size": preview_size,
                                            "preview_crop": preview_crop,
                                            "fields":       fields}, **kwargs)

    def process_args(self,
                     offset: int,
                     limit: int,
                     media_type: Optional[Union[str, Iterable[str]]],
                     sort: Optional[str],
                     preview_size: Optional[str],
                     preview_crop: Optional[bool],
                     fields: Optional[Fields]) -> None:
        self.params["offset"] = offset
        self.params["limit"] = limit

        if media_type is not None:
            if not isinstance(media_type, Iterable):
                raise TypeError("media_type should be a string or an iterable")

            if isinstance(media_type, str):
                self.params["media_type"] = media_type
            else:
                self.params["media_type"] = ",".join(media_type)

        if preview_size is not None:
            self.params["preview_size"] = preview_size

        if preview_crop is not None:
            self.params["preview_crop"] = preview_crop

        if sort is not None:
            self.params["sort"] = sort

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> FilesResourceListObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return FilesResourceListObject(js, yadisk)

class PatchRequest(APIRequest):
    """
        A request to update custom properties of a resource.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param path: path to the resource
        :param properties: `dict`, custom properties to update
        :param fields: list of keys to be included in the response

        :returns: :any:`ResourceObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk/resources"
    method = "PATCH"
    content_type = "application/json"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 path: str,
                 properties: dict,
                 fields: Optional[Fields] = None, **kwargs):
        APIRequest.__init__(self, session, {"path":       path,
                                            "properties": properties,
                                            "fields":     fields}, **kwargs)
    def process_args(self, path: str, properties: dict, fields: Optional[Fields]) -> None:
        self.params["path"] = ensure_path_has_schema(path)
        self.data = json.dumps({"custom_properties": properties}).encode("utf8")

        if fields is not None:
            sub_map = {"embedded": "_embedded"}

            self.params["fields"] = ",".join(_substitute_keys(fields, sub_map))

    def process_json(self,
                     js: Optional[dict],
                     yadisk: Optional["YaDisk"] = None) -> ResourceObject:
        return ResourceObject(js, yadisk)



# -*- coding: utf-8 -*-

from .api_request import APIRequest
from ..objects import DiskInfoObject
from ..exceptions import InvalidResponseError

from typing import Optional, TYPE_CHECKING
from ..compat import Iterable

if TYPE_CHECKING:
    import aiohttp

__all__ = ["DiskInfoRequest"]

class DiskInfoRequest(APIRequest):
    """
        A request to get disk information.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param fields: list of keys to be included in the response

        :returns: :any:`DiskInfoObject`
    """

    url = "https://cloud-api.yandex.net/v1/disk"
    method = "GET"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 fields: Optional[Iterable[str]] = None, **kwargs):
        APIRequest.__init__(self, session, {"fields": fields}, **kwargs)

    def process_args(self, fields: Optional[Iterable[str]]) -> None:
        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self, js: Optional[dict]) -> DiskInfoObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return DiskInfoObject(js)



# -*- coding: utf-8 -*-

import datetime

from .compat import Callable, List, AsyncIterable

from typing import Optional, TypeVar, Any, Union, IO, Protocol

__all__ = ["typed_list", "int_or_error", "str_or_error", "bool_or_error",
           "dict_or_error", "str_or_dict_or_error", "yandex_date", "is_operation_link",
           "is_resource_link", "is_public_resource_link", "ensure_path_has_schema",
           "FileOrPath", "FileOrPathDestination", "CaseInsensitiveDict"]

T = TypeVar("T", bound=Callable)

def typed_list(datatype: T) -> Callable[[Optional[List]], List[T]]:
    def list_factory(iterable: Optional[List] = None) -> List[T]:
        if iterable is None:
            return []

        if not isinstance(iterable, list):
            raise ValueError(f"Expected a list, got {type(iterable)}")

        return [datatype(i) for i in iterable]

    return list_factory

def int_or_error(x: Any) -> int:
    if not isinstance(x, int):
        raise ValueError(f"{repr(x)} is not an integer")

    return x

def str_or_error(x: Any) -> str:
    if not isinstance(x, str):
        raise ValueError(f"{repr(x)} is not a string")

    return x

def bool_or_error(x: Any) -> bool:
    if not isinstance(x, bool):
        raise ValueError(f"{repr(x)} is not a boolean value")

    return x

def dict_or_error(x: Any) -> dict:
    if not isinstance(x, dict):
        raise ValueError(f"{repr(x)} is not a dict")

    return x

def str_or_dict_or_error(x: Any) -> Union[str, dict]:
    if not isinstance(x, (str, dict)):
        raise ValueError(f"{repr(x)} is not a string nor a dict")

    return x

def yandex_date(string):
    return datetime.datetime.strptime(string[:-3] + string[-2:], "%Y-%m-%dT%H:%M:%S%z")

def is_operation_link(link: str) -> bool:
    if link.startswith("https://cloud-api.yandex.net/v1/disk/operations/"):
        return True

    # Same but http:// version
    return link.startswith("http://cloud-api.yandex.net/v1/disk/operations/")

def is_resource_link(url: str) -> bool:
    if url.startswith("https://cloud-api.yandex.net/v1/disk/resources?"):
        return True

    # Check also for HTTP version
    return url.startswith("http://cloud-api.yandex.net/v1/disk/resources?")

def is_public_resource_link(url: str) -> bool:
    if url.startswith("https://cloud-api.yandex.net/v1/disk/public/resources?"):
        return True

    # Check also for HTTP version
    return url.startswith("http://cloud-api.yandex.net/v1/disk/public/resources?")

def ensure_path_has_schema(path: str, default_schema: str = "disk") -> str:
    # Modifies path to always have a schema (disk:/, trash:/ or app:/).
    # Without the schema Yandex.Disk won't let you upload filenames with the ':' character.
    # See https://github.com/ivknv/yadisk/issues/26 for more details

    KNOWN_SCHEMAS = ("disk:", "trash:", "app:")

    if path in KNOWN_SCHEMAS:
        return default_schema + ":/" + path

    if path.startswith("/"):
        return default_schema + ":" + path

    if any(path.startswith(schema + "/") for schema in KNOWN_SCHEMAS):
        return path

    return default_schema + ":/" + path

class AsyncFileLike(Protocol):
    async def read(self, size: int = ..., /) -> Union[str, bytes]: ...
    async def write(self, buffer: Any, /) -> int: ...
    async def seek(self, pos: int, whence: int = ..., /) -> int: ...
    async def tell(self) -> int: ...

class BinaryAsyncFileLike(Protocol):
    async def read(self, size: int = ..., /) -> bytes: ...
    async def write(self, buffer: Any, /) -> int: ...
    async def seek(self, pos: int, whence: int = ..., /) -> int: ...
    async def tell(self) -> int: ...

FileOrPath = Union[
    str,
    bytes,
    IO,
    AsyncFileLike,
    Callable[[], AsyncIterable[bytes]]]

FileOrPathDestination = Union[
    str,
    bytes,
    IO[bytes],
    BinaryAsyncFileLike]

# https://stackoverflow.com/a/32888599/3653520
class CaseInsensitiveDict(dict):
    K = TypeVar("K")

    @classmethod
    def _k(cls, key: K) -> K:
        return key.lower() if isinstance(key, str) else key

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._convert_keys()

    def __getitem__(self, key: K) -> Any:
        return super().__getitem__(self.__class__._k(key))

    def __setitem__(self, key: K, value: Any) -> None:
        super().__setitem__(self.__class__._k(key), value)

    def __delitem__(self, key: K) -> Any:
        return super().__delitem__(self.__class__._k(key))

    def __contains__(self, key: K) -> bool:
        return super().__contains__(self.__class__._k(key))

    def pop(self, key: K, /, *args, **kwargs) -> Any:
        return super().pop(self.__class__._k(key), *args, **kwargs)

    def get(self, key: K, /, *args, **kwargs) -> Any:
        return super().get(self.__class__._k(key), *args, **kwargs)

    def setdefault(self, key: K, *args, **kwargs) -> Any:
        return super().setdefault(self.__class__._k(key), *args, **kwargs)

    def update(self, E: dict = {}, **F) -> None:
        super().update(self.__class__(E))
        super().update(self.__class__(**F))

    def _convert_keys(self) -> None:
        for k in list(self.keys()):
            v = super(CaseInsensitiveDict, self).pop(k)
            self.__setitem__(k, v)



# -*- coding: utf-8 -*-

from functools import partial

from .yadisk_object import YaDiskObject
from .resources import LinkObject
from ..common import str_or_error, dict_or_error

from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ..yadisk import YaDisk

__all__ = ["OperationStatusObject"]

class OperationStatusObject(YaDiskObject):
    """
        Operation status object.

        :param operation_status: `dict` or `None`
        :param yadisk: :any:`YaDisk` or `None`, `YaDisk` object

        :ivar type: `str`, type of the operation
        :ivar status: `str`, status of the operation
        :ivar operation_id: `str`, ID of the operation
        :ivar link: :any:`LinkObject`, link to the operation
        :ivar data: `dict`, other information about the operation
    """

    def __init__(self,
                 operation_status: Optional[dict] = None,
                 yadisk: Optional["YaDisk"] = None):
        YaDiskObject.__init__(
            self,
            {"type":         str_or_error,
             "status":       str_or_error,
             "operation_id": str_or_error,
             "link":         partial(LinkObject, yadisk=yadisk),
             "data":         dict_or_error},
            yadisk)

        self.import_fields(operation_status)



# -*- coding: utf-8 -*-

__all__ = ["List", "Dict", "Set", "Callable", "Iterable", "Generator",
           "AsyncGenerator", "Coroutine", "Awaitable", "TimeoutError", "AsyncIterable"]

import sys

if sys.version_info.major == 3 and sys.version_info.minor < 9:
    from typing import (
        List, Dict, Set, Callable, Iterable, Generator, AsyncGenerator,
        Coroutine, Awaitable, AsyncIterable
    )
else:
    from collections.abc import (
        Callable, Iterable, Generator, AsyncGenerator, Coroutine, Awaitable,
        AsyncIterable
    )

    List = list
    Dict = dict
    Set = set

if sys.version_info.major == 3 and sys.version_info.minor < 11:
    from asyncio import TimeoutError
else:
    TimeoutError = TimeoutError



# -*- coding: utf-8 -*-

import sys
import aiohttp

from .common import CaseInsensitiveDict

__all__ = ["SessionWithHeaders"]

DEFAULT_USER_AGENT = "Python/%s.%s aiohttp/%s" % (sys.version_info.major,
                                                  sys.version_info.minor,
                                                  aiohttp.__version__)


class SessionWithHeaders(aiohttp.ClientSession):
    """Just like your regular :any:`aiohttp.ClientSession` but with headers"""

    def __init__(self, *args, **kwargs):
        aiohttp.ClientSession.__init__(self, *args, **kwargs)

        self.headers.update(CaseInsensitiveDict({
            "User-Agent": DEFAULT_USER_AGENT,
            "Accept-Encoding": ", ".join(("gzip", "deflate")),
            "Accept": "*/*",
            "Connection": "keep-alive"
        }))




# -*- coding: utf-8 -*-

from functools import partial
from .yadisk_object import YaDiskObject
from ..common import str_or_error, bool_or_error, int_or_error

from typing import Optional, NoReturn, TYPE_CHECKING

if TYPE_CHECKING:
    from ..yadisk import YaDisk

__all__ = ["DiskInfoObject", "SystemFoldersObject", "UserObject", "UserPublicInfoObject"]

class DiskInfoObject(YaDiskObject):
    """
        Disk information object.

        :param disk_info: `dict` or `None`
        :param yadisk: :any:`YaDisk` or `None`, `YaDisk` object

        :ivar max_file_size: `int`, maximum supported file size (bytes)
        :ivar paid_max_file_size: `int`, maximum supported file size for a paid account (bytes)
        :ivar unlimited_autoupload_enabled: `bool`, tells whether unlimited
                                             autoupload from mobile devices is enabled
        :ivar total_space: `int`, total disk size (bytes)
        :ivar trash_size: `int`, amount of space used by trash (bytes), part of `used_space`
        :ivar is_paid: `bool`, tells if the account is paid or not
        :ivar used_space: `int`, amount of space used (bytes)
        :ivar system_folders: :any:`SystemFoldersObject`, paths to the system folders
        :ivar user: :any:`UserObject`, owner of the disk
        :ivar revision: `int`, current revision of Yandex.Disk
    """

    max_file_size:                Optional[int]
    paid_max_file_size:           Optional[int]
    unlimited_autoupload_enabled: Optional[bool]
    total_space:                  Optional[int]
    trash_size:                   Optional[int]
    is_paid:                      Optional[bool]
    used_space:                   Optional[int]
    system_folders:               "SystemFoldersObject"
    user:                         "UserObject"
    revision:                     Optional[int]

    def __init__(self, disk_info: Optional[dict] = None, yadisk: Optional["YaDisk"] = None):
        YaDiskObject.__init__(
            self,
            {"max_file_size":                int_or_error,
             "paid_max_file_size":           int_or_error,
             "unlimited_autoupload_enabled": bool_or_error,
             "total_space":                  int_or_error,
             "trash_size":                   int_or_error,
             "is_paid":                      bool_or_error,
             "used_space":                   int_or_error,
             "system_folders":               partial(SystemFoldersObject, yadisk=yadisk),
             "user":                         partial(UserObject, yadisk=yadisk),
             "revision":                     int_or_error},
            yadisk)

        self.import_fields(disk_info)

class SystemFoldersObject(YaDiskObject):
    """
        Object, containing paths to system folders.

        :param system_folders: `dict` or `None`
        :param yadisk: :any:`YaDisk` or `None`, `YaDisk` object

        :ivar odnoklassniki: `str`, path to the Odnoklassniki folder
        :ivar google: `str`, path to the Google+ folder
        :ivar instagram: `str`, path to the Instagram folder
        :ivar vkontakte: `str`, path to the VKontakte folder
        :ivar attach: `str`, path to the mail attachments folder
        :ivar mailru: `str`, path to the My World folder
        :ivar downloads: `str`, path to the Downloads folder
        :ivar applications: `str` path to the Applications folder
        :ivar facebook: `str`, path to the Facebook folder
        :ivar social: `str`, path to the social networks folder
        :ivar messenger: `str`, path to the Messenger Files folder
        :ivar calendar: `str`, path to the Meeting Materials folder
        :ivar photostream: `str`, path to the camera folder
        :ivar screenshots: `str`, path to the screenshot folder
        :ivar scans: `str`, path to the Scans folder
    """

    odnoklassniki: Optional[str]
    google:        Optional[str]
    instagram:     Optional[str]
    vkontakte:     Optional[str]
    attach:        Optional[str]
    mailru:        Optional[str]
    downloads:     Optional[str]
    applications:  Optional[str]
    facebook:      Optional[str]
    social:        Optional[str]
    messenger:     Optional[str]
    calendar:      Optional[str]
    photostream:   Optional[str]
    screenshots:   Optional[str]
    scans:         Optional[str]

    def __init__(self,
                 system_folders: Optional[dict] = None,
                 yadisk: Optional["YaDisk"] = None):
        YaDiskObject.__init__(
            self,
            {"odnoklassniki": str_or_error,
             "google":        str_or_error,
             "instagram":     str_or_error,
             "vkontakte":     str_or_error,
             "attach":        str_or_error,
             "mailru":        str_or_error,
             "downloads":     str_or_error,
             "applications":  str_or_error,
             "facebook":      str_or_error,
             "social":        str_or_error,
             "messenger":     str_or_error,
             "calendar":      str_or_error,
             "photostream":   str_or_error,
             "screenshots":   str_or_error,
             "scans":         str_or_error},
            yadisk)

        self.import_fields(system_folders)

class UserObject(YaDiskObject):
    """
        User object.

        :param user: `dict` or `None`
        :param yadisk: :any:`YaDisk` or `None`, `YaDisk` object

        :ivar country: `str`, user's country
        :ivar login: `str`, user's login
        :ivar display_name: `str`, user's display name
        :ivar uid: `str`, user's UID
    """

    country: Optional[str]
    login: Optional[str]
    display_name: Optional[str]
    uid: Optional[str]

    def __init__(self, user: Optional[dict] = None, yadisk: Optional["YaDisk"] = None):
        YaDiskObject.__init__(
            self,
            {"country":      str_or_error,
             "login":        str_or_error,
             "display_name": str_or_error,
             "uid":          str_or_error},
            yadisk)

        self.import_fields(user)

class UserPublicInfoObject(UserObject):
    """
        Public user information object.
        Inherits from :any:`UserObject` for compatibility.

        :param public_user_info: `dict` or `None`
        :param yadisk: :any:`YaDisk` or `None`, `YaDisk` object

        :ivar login: `str`, user's login
        :ivar display_name: `str`, user's display name
        :ivar uid: `str`, user's UID
    """

    country: NoReturn

    def __init__(self,
                 public_user_info: Optional[dict] = None,
                 yadisk: Optional["YaDisk"] = None):
        UserObject.__init__(self, None, yadisk)
        self.remove_field("country")
        self.import_fields(public_user_info)



# -*- coding: utf-8 -*-

from urllib.parse import urlparse, parse_qs, quote

from .api_request import APIRequest
from ..objects import OperationStatusObject
from ..common import is_operation_link
from ..exceptions import InvalidResponseError

from typing import Optional, TYPE_CHECKING
from ..compat import Iterable

if TYPE_CHECKING:
    import aiohttp

__all__ = ["GetOperationStatusRequest"]

class GetOperationStatusRequest(APIRequest):
    """
        A request to get operation status.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param operation_id: operation ID or link
        :param fields: list of keys to be included in the response

        :returns: :any:`OperationStatusObject`
    """

    method = "GET"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 operation_id: str,
                 fields: Optional[Iterable[str]] = None, **kwargs):
        if is_operation_link(operation_id):
            parsed_url = urlparse(operation_id)
            self.url = "https://" + parsed_url.netloc + parsed_url.path

            params = parse_qs(parsed_url.query)

            if fields is None:
                fields = params.get("fields", [None])[0]
        else:
            operation_id = quote(operation_id)
            self.url = "https://cloud-api.yandex.net/v1/disk/operations/%s" % (operation_id,)

        APIRequest.__init__(self, session, {"fields": fields}, **kwargs)

    def process_args(self, fields: Optional[Iterable[str]]) -> None:
        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self, js: Optional[dict]) -> OperationStatusObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        if "items" in js and isinstance(js["items"], list):
            if js["items"]:
                return OperationStatusObject(js["items"][0])

        return OperationStatusObject(js)



# -*- coding: utf-8 -*-

from .api_request import APIRequest
from ..objects import TokenObject, TokenRevokeStatusObject
from ..exceptions import InvalidResponseError

from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    import aiohttp

__all__ = ["RefreshTokenRequest", "RevokeTokenRequest", "GetTokenRequest"]

class RefreshTokenRequest(APIRequest):
    """
        A request to refresh an existing token.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param refresh_token: the refresh token that was received with the original token
        :param client_id: application ID
        :param client_secret: application secret password

        :returns: :any:`TokenObject`
    """

    url = "https://oauth.yandex.ru/token"
    method = "POST"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 refresh_token: str,
                 client_id: str,
                 client_secret: str, **kwargs):
        APIRequest.__init__(self, session, {"refresh_token": refresh_token,
                                            "client_id":     client_id,
                                            "client_secret": client_secret}, **kwargs)

    def process_args(self, refresh_token: str, client_id: str, client_secret: str) -> None:
        self.data["grant_type"] = "refresh_token"
        self.data["refresh_token"] = refresh_token
        self.data["client_id"] = client_id
        self.data["client_secret"] = client_secret

    def process_json(self, js: Optional[dict]) -> TokenObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk did not return valid JSON")

        return TokenObject(js)

class RevokeTokenRequest(APIRequest):
    """
        A request to revoke the token.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param token: the token to be revoked
        :param client_id: application ID
        :param client_secret: application secret password

        :returns: :any:`TokenRevokeStatusObject`
    """

    url = "https://oauth.yandex.ru/revoke_token"
    method = "POST"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 token: str,
                 client_id: str,
                 client_secret: str, **kwargs):
        APIRequest.__init__(self, session, {"token":         token,
                                            "client_id":     client_id,
                                            "client_secret": client_secret}, **kwargs)

    def process_args(self, token: str, client_id: str, client_secret: str) -> None:
        self.data["access_token"] = token
        self.data["client_id"] = client_id
        self.data["client_secret"] = client_secret

    def process_json(self, js: Optional[dict]) -> TokenRevokeStatusObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk did not return valid JSON")

        return TokenRevokeStatusObject(js)

class GetTokenRequest(APIRequest):
    """
        A request to get the token.

        :param session: an instance of :any:`aiohttp.ClientSession` with prepared headers
        :param code: confirmation code
        :param client_id: application ID
        :param client_secret: application secret password
        :param device_id: unique device ID (between 6 and 50 characters)

        :returns: :any:`TokenObject`
    """

    url = "https://oauth.yandex.ru/token"
    method = "POST"

    def __init__(self,
                 session: "aiohttp.ClientSession",
                 code: str,
                 client_id: str,
                 client_secret: str,
                 device_id: Optional[str] = None,
                 device_name: Optional[str] = None, **kwargs):
        APIRequest.__init__(self, session, {"code":          code,
                                            "client_id":     client_id,
                                            "client_secret": client_secret,
                                            "device_id":     device_id,
                                            "device_name":   device_name}, **kwargs)

    def process_args(self,
                     code: str,
                     client_id: str,
                     client_secret: str,
                     device_id: Optional[str],
                     device_name: Optional[str]) -> None:
        self.data["grant_type"] = "authorization_code"
        self.data["code"] = code
        self.data["client_id"] = client_id
        self.data["client_secret"] = client_secret

        if device_id is not None:
            self.data["device_id"] = device_id

        if device_name is not None:
            self.data["device_name"] = device_name

    def process_json(self, js: Optional[dict]) -> TokenObject:
        if js is None:
            raise InvalidResponseError("Yandex.Disk did not return valid JSON")

        return TokenObject(js)



# -*- coding: utf-8 -*-

from . import api, objects, exceptions, utils
from .yadisk import YaDisk

import warnings

__version__ = "1.4.5"

warnings.warn("This library is no longer maintained, use yadisk>=2.0.0 instead, it supports async API", DeprecationWarning)
